-------------------------------------------------------------------------------
-- $Id$
-------------------------------------------------------------------------------
-- do_the_mult.vhd - Entity and architecture
--
--
-------------------------------------------------------------------------------
-- Filename:        do_the_mult.vhd
--
-- Description:     
--                  
-- VHDL-Standard:   VHDL'93
-------------------------------------------------------------------------------
-- Structure:   
--              do_the_mult.vhd
--
-------------------------------------------------------------------------------
-- Author:          
-- Revision:        $Revision$
-- Date:            $Date$
--
-- History:
--    2003-11-12    First Version
--
-------------------------------------------------------------------------------
-- Naming Conventions:
--      active low signals:                     "*_n"
--      clock signals:                          "clk", "clk_div#", "clk_#x" 
--      reset signals:                          "rst", "rst_n" 
--      generics:                               "C_*" 
--      user defined types:                     "*_TYPE" 
--      state machine next state:               "*_ns" 
--      state machine current state:            "*_cs" 
--      combinatorial signals:                  "*_com" 
--      pipelined or register delay signals:    "*_d#" 
--      counter signals:                        "*cnt*"
--      clock enable signals:                   "*_ce" 
--      internal version of output port         "*_i"
--      device pins:                            "*_pin" 
--      ports:                                  - Names begin with Uppercase 
--      processes:                              "*_PROCESS" 
--      component instantiations:               "<ENTITY_>I_<#|FUNC>
-------------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;

entity do_the_mult is
  port (
    Clk         : in  std_logic;
    Reset       : in  std_logic;
    CE          : in  std_logic;
    Valid_Input : in  std_logic;
    Op1         : in  std_logic_vector(0 to 31);
    Op2         : in  std_logic_vector(0 to 31);
    Valid_Res   : out std_logic;
    Result      : out std_logic_vector(0 to 31));
end entity do_the_mult;

library unisim;
use unisim.all;

architecture IMP of do_the_mult is

  component MULT18x18S is
    port (
      A  : in  std_logic_vector(17 downto 0);
      B  : in  std_logic_vector(17 downto 0);
      C  : in  std_logic;
      CE : in  std_logic;
      R  : in  std_logic;
      P  : out std_logic_vector(35 downto 0)
      );
  end component MULT18x18S;

  component FDR is
    port (
      Q : out std_logic;
      C : in  std_logic;
      D : in  std_logic;
      R : in  std_logic);
  end component FDR;

-------------------------------------------------------------------------------
-- Begin architecture
-------------------------------------------------------------------------------

  signal a_Oper : std_logic_vector(0 to 17);
  signal b_Oper : std_logic_vector(0 to 17);
  signal C_OPER : std_logic_vector(0 to 17);
  signal d_Oper : std_logic_vector(0 to 17);

  signal stage_valid  : std_logic;
  signal stage1_valid : std_logic;
  signal stage2_valid : std_logic;
  signal stage3_valid : std_logic;
  signal stage4_valid : std_logic;

  signal sign_A    : std_logic;
  signal sign_C    : std_logic;
  signal prod_BD   : std_logic_vector(0 to 35);
  signal prod_AD   : std_logic_vector(0 to 35);
  signal prod_BC   : std_logic_vector(0 to 35);
  signal prod_AC   : std_logic_vector(0 to 35);
  signal prod_BD_1 : std_logic_vector(0 to 31);
  signal prod_AD_1 : std_logic_vector(0 to 31);
  signal prod_BC_1 : std_logic_vector(0 to 31);
  signal prod_AC_1 : std_logic_vector(0 to 31);

  signal stage_sum     : std_logic_vector(0 to 31);
  signal Stage_Sum_1   : std_logic_vector(0 to 31);
  signal Stage_Sum_2_i : std_logic_vector(0 to 31);
  signal Stage_Sum_2   : std_logic_vector(0 to 31);
  signal Stage_Sum_3_i : std_logic_vector(0 to 31);
  signal Stage_Sum_3   : std_logic_vector(0 to 31);

  signal prod_ac_2 : std_logic_vector(0 to 31);
  signal prod_ac_3 : std_logic_vector(0 to 31);
  signal prod_bc_2 : std_logic_vector(0 to 31);

begin  -- architecture IMP


  -----------------------------------------------------------------------------
  -----------------------------------------------------------------------------
  -- Perform (32x32=64) >> 28
  -- Operand1 = AB
  -- Operand2 = CD
  -----------------------------------------------------------------------------
  -----------------------------------------------------------------------------

  sign_A <= Op1(Op1'left);
  sign_C <= Op2(Op2'left);

  A_Oper <= sign_A & sign_A & Op1(0 to 15);
  B_Oper <= "00" & Op1(16 to 31);
  C_OPER <= sign_C & sign_C & Op2(0 to 15);
  D_Oper <= "00" & Op2(16 to 31);

  MULT18x18_BD : MULT18x18S
    port map (
      A  => b_Oper,                     -- [in]
      B  => d_Oper,                     -- [in]
      C  => Clk,                        -- [in]
      CE => CE,                         -- [in]
      R  => Reset,                      -- [in]
      P  => prod_BD);                   -- [out]

  MULT18x18_AD : MULT18x18S
    port map (
      A  => a_Oper,                     -- [in]
      B  => d_Oper,                     -- [in]
      C  => Clk,                        -- [in]
      CE => CE,                         -- [in]
      R  => Reset,                      -- [in]
      P  => prod_AD);                   -- [out]

  MULT18x18_BC : MULT18x18S
    port map (
      A  => C_OPER,                     -- [in]
      B  => b_Oper,                     -- [in]
      C  => Clk,                        -- [in]
      CE => CE,                         -- [in]
      R  => Reset,                      -- [in]
      P  => prod_BC);                   -- [out]

  MULT18x18_AC : MULT18x18S
    port map (
      A  => A_OPER,                     -- [in]
      B  => C_OPER,                     -- [in]
      C  => Clk,                        -- [in]
      CE => CE,                         -- [in]
      R  => Reset,                      -- [in]
      P  => prod_AC);                   -- [out]

  Valid_DFFs : process (Clk, Reset) is
  begin  -- process Valid_DFFs
    if Reset = '1' then                 -- asynchronous reset (active high)
      stage_valid <= '0';
    elsif Clk'event and Clk = '1' then  -- rising clock edge
      if (CE = '1') then
        stage_valid <= Valid_Input;
      end if;
    end if;
  end process Valid_DFFs;



  -------------------------------------------------------------------------
  -- Need to insert DFF to the output from the multipliers.
  -- Otherwise a 100 MHz clock frequency can't be maintainded
  -------------------------------------------------------------------------
  Mult_DFFs : process (Clk, Reset) is
  begin  -- process Mult_DFFs
    if Reset = '1' then                 -- asynchronous reset (active high)
      prod_AC_1    <= (others => '0');
      prod_AD_1    <= (others => '0');
      prod_BC_1    <= (others => '0');
      prod_BD_1    <= (others => '0');
      stage1_valid <= '0';
    elsif Clk'event and Clk = '1' then  -- rising clock edge
      if (CE = '1') then
        stage1_valid <= stage_valid;
        prod_BD_1    <= prod_BD(4 to 35);
        prod_AD_1    <= prod_AD(4 to 35);
        prod_BC_1    <= prod_BC(4 to 35);
        prod_AC_1    <= prod_AC(4 to 35);
      end if;
    end if;

  end process Mult_DFFs;

  -- Reduce the prod_BD since we don't use the least 16 bits of the result
  -- Also align with the AD vector
--  prod_BD_i(16 to 31) <= prod_BD(4 to to 35);

  Stage_Sum <= std_logic_vector(unsigned(prod_AD_1(0 to 31)) + unsigned(prod_BD_1(0 to 15)));

  -----------------------------------------------------------------------------
  -- Stage 3 of flipflops
  -----------------------------------------------------------------------------
  Stage_2_DFFs : process (Clk, Reset) is
  begin  -- process Stage_2_DFFs
    if Reset = '1' then                 -- asynchronous reset (active high)
      stage2_valid <= '0';
      Stage_Sum_1  <= (others => '0');
      prod_AC_2    <= (others => '0');
      prod_BC_2    <= (others => '0');
    elsif Clk'event and Clk = '1' then  -- rising clock edge
      if (CE = '1') then
        stage2_valid <= stage1_valid;
        Stage_Sum_1  <= Stage_Sum;
        prod_AC_2    <= prod_AC_1(0 to 31);
        prod_BC_2    <= prod_BC_1(0 to 31);
      end if;
    end if;
  end process Stage_2_DFFs;

  -- Stage 2 of adders
  Stage_Sum_2_i <= std_logic_vector(signed(Stage_Sum_1) + signed(prod_BC_2));

  -----------------------------------------------------------------------------
  -- Stage 3 of flipflops
  -----------------------------------------------------------------------------
  Stage_3_DFFs : process (Clk, Reset) is
  begin  -- process Stage_3_DFFs
    if Reset = '1' then                 -- asynchronous reset (active high)
      Stage_Sum_2  <= (others => '0');
      prod_AC_3    <= (others => '0');
      stage3_valid <= '0';
    elsif Clk'event and Clk = '1' then  -- rising clock edge
      if (CE = '1') then
        stage3_valid <= stage2_valid;
        Stage_Sum_2  <= Stage_Sum_2_i;
        prod_AC_3    <= prod_AC_2;
      end if;
    end if;
  end process Stage_3_DFFs;

  -- Stage 3 of adders
  Stage_Sum_3_i <= std_logic_vector(signed(Stage_Sum_2(0 to 15)) + signed(prod_AC_3));

  -----------------------------------------------------------------------------
  -- Stage 4 of flipflops
  -----------------------------------------------------------------------------
  Stage_4_DFFs : process (Clk, Reset) is
  begin  -- process Stage_4_DFFs
    if Reset = '1' then                 -- asynchronous reset (active high)
      Stage_Sum_3  <= (others => '0');
      Stage4_valid <= '0';
    elsif Clk'event and Clk = '1' then  -- rising clock edge
      Stage4_valid <= '0';
      if (CE = '1') then
        stage4_valid <= stage3_valid;
        Stage_Sum_3  <= Stage_Sum_3_i(4 to 31) & Stage_Sum_2(16 to 19);
      end if;
    end if;
  end process Stage_4_DFFs;

  Valid_Res <= stage4_valid;

  result <= Stage_Sum_3;

end architecture IMP;
